@@ -1,15 +1,32 @@
-from typing import List, Dict, Any
+from typing import List, Dict, Optional
+import time
+from dataclasses import dataclass
 import backoff  # type: ignore
 import requests
 from requests.exceptions import Timeout
 import spectacles.utils as utils
+from spectacles.types import JsonDict
 from spectacles.logger import GLOBAL_LOGGER as logger
 from spectacles.exceptions import SpectaclesException, LookerApiError
 
-JsonDict = Dict[str, Any]
 TIMEOUT_SEC = 300
 
 
+@dataclass(frozen=True)  # Token is immutable
+class AccessToken:
+    access_token: str
+    token_type: str
+    expires_in: int
+    expires_at: float
+
+    def __str__(self) -> str:
+        return self.access_token
+
+    @property
+    def expired(self) -> bool:
+        return False if time.time() < self.expires_at else True
+
+
 class LookerClient:
     """Wraps some endpoints of the Looker API, issues requests and handles responses.
 
@@ -48,13 +65,15 @@ class LookerClient:
 
         self.base_url: str = base_url.rstrip("/")
         self.api_url: str = f"{self.base_url}:{port}/api/{api_version}/"
+        self.client_id: str = client_id
+        self.client_secret: str = client_secret
+        self.api_version: float = api_version
+        self.access_token: Optional[AccessToken] = None
         self.session: requests.Session = requests.Session()
 
-        self.authenticate(client_id, client_secret, api_version)
+        self.authenticate()
 
-    def authenticate(
-        self, client_id: str, client_secret: str, api_version: float
-    ) -> None:
+    def authenticate(self) -> None:
         """Logs in to Looker's API using a client ID/secret pair and an API version.
 
         Args:
@@ -66,7 +85,8 @@ class LookerClient:
         logger.debug("Authenticating Looker API credentials")
 
         url = utils.compose_url(self.api_url, path=["login"])
-        body = {"client_id": client_id, "client_secret": client_secret}
+        body = {"client_id": self.client_id, "client_secret": self.client_secret}
+        # This should not use `self.post` or it will create a recursive loop
         response = self.session.post(url=url, data=body, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
@@ -76,21 +96,48 @@ class LookerClient:
                 title="Couldn't authenticate to the Looker API.",
                 status=response.status_code,
                 detail=(
-                    f"Unable to authenticate with client ID '{client_id}'. "
+                    f"Unable to authenticate with client ID '{self.client_id}'. "
                     "Check that your credentials are correct and try again."
                 ),
                 response=response,
             )
 
-        access_token = response.json()["access_token"]
-        self.session.headers = {"Authorization": f"token {access_token}"}
+        result = response.json()
+        if "expires_at" not in result:
+            # Calculate the expiration time with a one-minute buffer
+            result["expires_at"] = time.time() + result["expires_in"] - 60
+        self.access_token = AccessToken(**result)
+        self.session.headers = {  # type: ignore
+            "Authorization": f"token {self.access_token}"
+        }
 
         looker_version = self.get_looker_release_version()
         logger.info(
             f"Connected to Looker version {looker_version} "
-            f"using Looker API {api_version}"
+            f"using Looker API {self.api_version}"
         )
 
+    def request(self, method: str, url: str, *args, **kwargs) -> requests.Response:
+        if self.access_token and self.access_token.expired:
+            logger.debug("Looker API access token has expired, requesting a new one")
+            self.authenticate()
+        return self.session.request(method, url, *args, **kwargs)
+
+    def get(self, url, *args, **kwargs) -> requests.Response:
+        return self.request("GET", url, *args, **kwargs)
+
+    def post(self, url, *args, **kwargs) -> requests.Response:
+        return self.request("POST", url, *args, **kwargs)
+
+    def patch(self, url, *args, **kwargs) -> requests.Response:
+        return self.request("PATCH", url, *args, **kwargs)
+
+    def put(self, url, *args, **kwargs) -> requests.Response:
+        return self.request("PUT", url, *args, **kwargs)
+
+    def delete(self, url, *args, **kwargs) -> requests.Response:
+        return self.request("DELETE", url, *args, **kwargs)
+
     def get_looker_release_version(self) -> str:
         """Gets the version number of connected Looker instance.
 
@@ -102,7 +149,7 @@ class LookerClient:
 
         url = utils.compose_url(self.api_url, path=["versions"])
 
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -111,7 +158,7 @@ class LookerClient:
                 title="Couldn't get Looker's release version.",
                 status=response.status_code,
                 detail=(
-                    f"Unable to get the release version of your Looker instance. "
+                    "Unable to get the release version of your Looker instance. "
                     "Please try again."
                 ),
                 response=response,
@@ -129,7 +176,7 @@ class LookerClient:
         logger.debug(f"Updating session to use the {workspace} workspace")
         url = utils.compose_url(self.api_url, path=["session"])
         body = {"workspace_id": workspace}
-        response = self.session.patch(url=url, json=body, timeout=TIMEOUT_SEC)
+        response = self.patch(url=url, json=body, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -156,7 +203,7 @@ class LookerClient:
         url = utils.compose_url(
             self.api_url, path=["projects", project, "git_branches"]
         )
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -183,7 +230,7 @@ class LookerClient:
         logger.debug(f"Setting Git branch to '{branch}'")
         url = utils.compose_url(self.api_url, path=["projects", project, "git_branch"])
         body = {"name": branch}
-        response = self.session.put(url=url, json=body, timeout=TIMEOUT_SEC)
+        response = self.put(url=url, json=body, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -206,11 +253,11 @@ class LookerClient:
             project: Name of the Looker project to use.
 
         """
-        logger.debug(f"Resetting branch to remote.")
+        logger.debug("Resetting branch to remote.")
         url = utils.compose_url(
             self.api_url, path=["projects", project, "reset_to_remote"]
         )
-        response = self.session.post(url=url, timeout=TIMEOUT_SEC)
+        response = self.post(url=url, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -219,7 +266,7 @@ class LookerClient:
                 title="Couldn't checkout Git branch.",
                 status=response.status_code,
                 detail=(
-                    f"Unable to reset local Git branch"
+                    "Unable to reset local Git branch"
                     "to match remote. Please try again."
                 ),
                 response=response,
@@ -234,9 +281,9 @@ class LookerClient:
         Returns:
             List[JsonDict]: JSON response containing all dependent projects
         """
-        logger.debug(f"Getting manifest details")
+        logger.debug("Getting manifest details")
         url = utils.compose_url(self.api_url, path=["projects", project, "manifest"])
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -268,7 +315,7 @@ class LookerClient:
         """
         logger.debug(f"Getting active branch for project '{project}'")
         url = utils.compose_url(self.api_url, path=["projects", project, "git_branch"])
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -279,7 +326,7 @@ class LookerClient:
                 status=response.status_code,
                 detail=(
                     f"Unable to get active branch for project '{project}'. "
-                    "Please try again."
+                    "Please check that the project exists and try again."
                 ),
                 response=response,
             )
@@ -308,7 +355,7 @@ class LookerClient:
 
         body = {"name": branch, "ref": ref}
         url = utils.compose_url(self.api_url, path=["projects", project, "git_branch"])
-        response = self.session.post(url=url, json=body, timeout=TIMEOUT_SEC)
+        response = self.post(url=url, json=body, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -339,7 +386,7 @@ class LookerClient:
 
         body = {"name": branch, "ref": ref}
         url = utils.compose_url(self.api_url, path=["projects", project, "git_branch"])
-        response = self.session.put(url=url, json=body, timeout=TIMEOUT_SEC)
+        response = self.put(url=url, json=body, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -368,7 +415,7 @@ class LookerClient:
         url = utils.compose_url(
             self.api_url, path=["projects", project, "git_branch", branch]
         )
-        response = self.session.delete(url=url, timeout=TIMEOUT_SEC)
+        response = self.delete(url=url, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -398,7 +445,7 @@ class LookerClient:
         url = utils.compose_url(
             self.api_url, path=["projects", project, "lookml_tests"]
         )
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -416,7 +463,9 @@ class LookerClient:
 
         return response.json()
 
-    def run_lookml_test(self, project: str, model: str = None) -> List[JsonDict]:
+    def run_lookml_test(
+        self, project: str, model: str = None, test: str = None
+    ) -> List[JsonDict]:
         """Runs all LookML/data tests for a given project and model (optional)
 
         This command only runs tests in production, as the Looker API doesn't currently
@@ -430,16 +479,25 @@ class LookerClient:
             List[JsonDict]: JSON response containing any LookML/data test errors
 
         """
-        logger.debug(f"Running LookML tests for project {project}")
+        if model is None and test is None:
+            logger.debug(f"Running all LookML tests for project '{project}'")
+        elif model is None and test is not None:
+            logger.debug(f"Running LookML test '{test}'")
+        elif model is not None and test is None:
+            logger.debug(f"Running all LookML tests for model '{model}'")
+        elif model is not None and test is not None:
+            logger.debug(f"Running LookML test '{test}' in model '{model}'")
+
         url = utils.compose_url(
             self.api_url, path=["projects", project, "lookml_tests", "run"]
         )
+
+        params = {}
         if model is not None:
-            response = self.session.get(
-                url=url, params={"model": model}, timeout=TIMEOUT_SEC
-            )
-        else:
-            response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+            params["model"] = model
+        if test is not None:
+            params["test"] = test
+        response = self.session.get(url=url, params=params, timeout=TIMEOUT_SEC)
 
         try:
             response.raise_for_status()
@@ -466,7 +524,7 @@ class LookerClient:
         """
         logger.debug(f"Getting all models and explores from {self.base_url}")
         url = utils.compose_url(self.api_url, path=["lookml_models"])
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -496,7 +554,7 @@ class LookerClient:
         url = utils.compose_url(
             self.api_url, path=["lookml_models", model, "explores", explore]
         )
-        response = self.session.get(url=url, timeout=TIMEOUT_SEC)
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -539,7 +597,7 @@ class LookerClient:
             "filter_expression": "1=2",
         }
         url = utils.compose_url(self.api_url, path=["queries"])
-        response = self.session.post(url=url, json=body, timeout=TIMEOUT_SEC)
+        response = self.post(url=url, json=body, timeout=TIMEOUT_SEC)
         try:
             response.raise_for_status()
         except requests.exceptions.HTTPError:
@@ -586,7 +644,7 @@ class LookerClient:
         body = {"query_id": query_id, "result_format": "json_detail"}
         url = utils.compose_url(self.api_url, path=["query_tasks"])
 
-        response = self.session.post(
+        response = self.post(
             url=url, json=body, params={"cache": "false"}, timeout=TIMEOUT_SEC
         )
 
@@ -626,7 +684,7 @@ class LookerClient:
             "Attempting to get results for %d query tasks", len(query_task_ids)
         )
         url = utils.compose_url(self.api_url, path=["query_tasks", "multi_results"])
-        response = self.session.get(
+        response = self.get(
             url=url,
             params={"query_task_ids": ",".join(query_task_ids)},
             timeout=TIMEOUT_SEC,
@@ -659,7 +717,45 @@ class LookerClient:
         """
         logger.debug(f"Cancelling query task: {query_task_id}")
         url = utils.compose_url(self.api_url, path=["running_queries", query_task_id])
-        self.session.delete(url=url, timeout=TIMEOUT_SEC)
+        self.delete(url=url, timeout=TIMEOUT_SEC)
 
         # No raise_for_status() here because Looker API seems to give a 404
         # if you try to cancel a finished query which can happen as part of cleanup
+
+    def content_validation(self) -> JsonDict:
+        logger.debug("Validating all content in Looker")
+        url = utils.compose_url(self.api_url, path=["content_validation"])
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
+
+        try:
+            response.raise_for_status()
+        except requests.exceptions.HTTPError:
+            raise LookerApiError(
+                name="unable-to-validate-content",
+                title="Couldn't validate Looks and Dashboards.",
+                status=response.status_code,
+                detail=("Failed to run the content validator. Please try again."),
+                response=response,
+            )
+
+        result = response.json()
+        return result
+
+    def all_folders(self, project: str) -> List[JsonDict]:
+        logger.debug("Getting information about all folders")
+        url = utils.compose_url(self.api_url, path=["folders"])
+        response = self.get(url=url, timeout=TIMEOUT_SEC)
+
+        try:
+            response.raise_for_status()
+        except requests.exceptions.HTTPError:
+            raise LookerApiError(
+                name="unable-to-get-folders",
+                title="Couldn't obtain project folders.",
+                status=response.status_code,
+                detail=(f"Failed to get all folders for project '{project}'."),
+                response=response,
+            )
+
+        result = response.json()
+        return result
