@@ -1,6 +1,7 @@
 import re
 from typing import Dict, List, Sequence, Optional, Any
-from spectacles.exceptions import SqlError
+from spectacles.exceptions import ValidationError
+from spectacles.types import QueryMode
 
 
 class LookMlObject:
@@ -25,7 +26,7 @@ class Dimension(LookMlObject):
         self.sql = sql
         self.url = url
         self.queried: bool = False
-        self.error: Optional[SqlError] = None
+        self.errors: List[ValidationError] = []
         if re.search(r"spectacles\s*:\s*ignore", sql, re.IGNORECASE):
             self.ignore = True
         else:
@@ -52,14 +53,14 @@ class Dimension(LookMlObject):
 
     @property
     def errored(self):
-        return bool(self.error) if self.queried else None
+        return bool(self.errors) if self.queried else None
 
     @errored.setter
     def errored(self, value):
         raise AttributeError(
             "Cannot assign to 'errored' property of a Dimension instance. "
-            "For a dimension to be considered errored, it must have a SqlError "
-            "in its 'error' attribute."
+            "For a dimension to be considered errored, it must have a ValidationError "
+            "in its 'errors' attribute."
         )
 
     @classmethod
@@ -76,8 +77,8 @@ class Explore(LookMlObject):
         self.name = name
         self.model_name = model_name
         self.dimensions = [] if dimensions is None else dimensions
-        self.queried: bool = False
-        self.error: Optional[SqlError] = None
+        self.errors: List[ValidationError] = []
+        self._queried: bool = False
 
     def __eq__(self, other):
         if not isinstance(other, Explore):
@@ -89,10 +90,27 @@ class Explore(LookMlObject):
             and self.dimensions == other.dimensions
         )
 
+    @property
+    def queried(self):
+        if self.dimensions:
+            return any(dimension.queried for dimension in self.dimensions)
+        else:
+            return self._queried
+
+    @queried.setter
+    def queried(self, value: bool):
+        if not isinstance(value, bool):
+            raise TypeError("Value for queried must be boolean.")
+        if self.dimensions:
+            for dimension in self.dimensions:
+                dimension.queried = value
+        else:
+            self._queried = value
+
     @property
     def errored(self):
         if self.queried:
-            return bool(self.error) or any(
+            return bool(self.errors) or any(
                 dimension.errored for dimension in self.dimensions
             )
         else:
@@ -102,22 +120,10 @@ class Explore(LookMlObject):
     def errored(self, value):
         raise AttributeError(
             "Cannot assign to 'errored' property of an Explore instance. "
-            "For an explore to be considered errored, it must have a SqlError "
-            "in its 'error' attribute or contain dimensions in an errored state."
+            "For an explore to be considered errored, it must have a ValidationError "
+            "in its 'errors' attribute or contain dimensions in an errored state."
         )
 
-    @property
-    def queried(self):
-        return any(dimension.queried for dimension in self.dimensions)
-
-    @queried.setter
-    def queried(self, value: bool):
-        if not isinstance(value, bool):
-            raise TypeError("Value for queried must be boolean.")
-        for dimension in self.dimensions:
-            if not dimension.ignore:
-                dimension.queried = value
-
     def get_errored_dimensions(self):
         for dimension in self.dimensions:
             if dimension.errored:
@@ -134,11 +140,13 @@ class Explore(LookMlObject):
     @property
     def number_of_errors(self):
         if self.errored:
-            if self.error:
-                errors = 1
+            if self.errors:
+                errors = len(self.errors)
             else:
-                errors = len(
-                    [dimension for dimension in self.dimensions if dimension.errored]
+                errors = sum(
+                    len(dimension.errors)
+                    for dimension in self.dimensions
+                    if dimension.errored
                 )
             return errors
         else:
@@ -193,6 +201,11 @@ class Model(LookMlObject):
         for explore in self.explores:
             explore.queried = value
 
+    def get_explore(self, name: str) -> Optional[Explore]:
+        return next(
+            (explore for explore in self.explores if explore.name == name), None
+        )
+
     def get_errored_explores(self):
         for explore in self.explores:
             if explore.errored:
@@ -225,6 +238,9 @@ class Project(LookMlObject):
 
         return self.name == other.name and self.models == other.models
 
+    def count_explores(self) -> int:
+        return sum(len(model.explores) for model in self.models)
+
     @property
     def errored(self):
         if self.queried:
@@ -259,17 +275,29 @@ class Project(LookMlObject):
             if model.errored:
                 yield model
 
-    def get_results(self) -> Dict[str, Any]:
-        errors = []
+    def get_model(self, name: str) -> Optional[Model]:
+        return next((model for model in self.models if model.name == name), None)
+
+    def get_explore(self, model: str, name: str) -> Optional[Explore]:
+        model_object = self.get_model(model)
+        if not model_object:
+            return None
+        else:
+            return model_object.get_explore(name)
+
+    def get_results(
+        self, validator: str, mode: Optional[QueryMode] = None
+    ) -> Dict[str, Any]:
+        errors: List[Dict[str, Any]] = []
         tested = []
 
         def parse_explore_errors(explore):
-            if explore.error:
-                errors.append(explore.error.__dict__)
+            if validator != "sql" or mode == "batch":
+                errors.extend([error.__dict__ for error in explore.errors])
             else:
                 for dimension in explore.dimensions:
                     if dimension.errored:
-                        errors.append(dimension.error.__dict__)
+                        errors.extend([error.__dict__ for error in dimension.errors])
 
         for model in self.models:
             for explore in model.explores:
@@ -286,7 +314,7 @@ class Project(LookMlObject):
 
         passed = min((test["passed"] for test in tested), default=True)
         return {
-            "validator": "sql",
+            "validator": validator,
             "status": "passed" if passed else "failed",
             "tested": tested,
             "errors": errors,
